/* wmb_libs.c - Edward H. Flora - ehf_dockapps@cox.net */
/* Last Modified: 4/3/04 */
/* 
 * These functions are designed to work with wmbutton.
 */

/* PORTABILITY:
******************
* Coded in ANSI C, using ANSI prototypes.
*/

/******  Include Files  *************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "wmbutton.h"


/******  ToolTip Globals  ***********************************************/

static struct timeval _tStart;

int _bTooltip = 0;
int _bTooltipSupport;
int _bTooltipSwapColors = 0;
XFontStruct* _fTooltip;
int _nFontHeight, _nFontY;
int _nScreenWidth, _nScreenHeight;
GC _gcMono= 0;
Window _wTooltip;

/******  Parse Command Line  ********************************************/
void parsargs(int argc, char **argv, char *Geometry_str, char *Display_str ) {
  extern int  Verbose;
  extern char configfile[FILEMAXLEN];
  extern char buttonfile[FILEMAXLEN];
  extern char szTooltipFont[TOOLTIP_FONT_LEN];
  extern int  bTooltipSwapColors;
  extern int  bTooltipSupport;
  extern int  mmouse;
  int i;

  /* Parse Command Line Arguments */  
  for ( i=1; i<argc; i++ ) {
    if ( *argv[i] == '-' ) {
      switch ( *(argv[i]+1) ) {
      case 'v':                        // Turn on Verbose (debugging) Mode
	Verbose = 1;
	break;
      case 'g':                        // Set Geometry Options
	if ( ++i >= argc ) show_usage();
	strncpy( Geometry_str, argv[i], sizeof( Geometry_str )-1);
	if ( Verbose ) printf("Geometry is: %s\n", Geometry_str);
	break;
      case 'd':                        // Set display 
	if ( ++i >= argc ) show_usage();
	strncpy( Display_str, argv[i], sizeof( Display_str )-1);
	if ( Verbose ) printf("Display is: %s\n", Display_str);
	break;
      case 'h':                        // Show Help Message
	show_usage();
	break;
      case 'F':                        // tooltip font <font>
	if ( ++i >= argc ) show_usage();
	strncpy( szTooltipFont, argv[i], sizeof( szTooltipFont )-1);
	if ( Verbose ) printf("Using Font: %s\n", szTooltipFont);
	break;
      case 's':                        // tooltip swap colors
	bTooltipSwapColors = 1;
	break;
      case 'n':                        // disable tooltips
	bTooltipSupport = 0;
	break;
      case 'f':                        // use config file <filename>
	if ( ++i >= argc ) show_usage();
	strncpy( configfile, argv[i], sizeof( configfile)-1);
	if ( Verbose ) printf("Using Config File: %s\n", configfile);
	break;
      case 'b':                        // use button file <filename>
	if ( ++i >= argc ) show_usage();
	strncpy( buttonfile, argv[i], sizeof( buttonfile)-1);
	if ( Verbose ) printf("Using Button File: %s\n", buttonfile);
	break;
      case 'm':                        // Enable Middle Mouse
	mmouse = !(MIDMOUSE);
	break;
      default:                         // other, unknown, parameters
	show_usage();
	break;
      }
    }
  }
}/***********************************************************************/

/******  Show Usage Information  ****************************************/
void show_usage() {
  extern char *app_name;

  fprintf(stderr,"\n");
  fprintf(stderr,"usage: %s [-g geom] [-d dpy] [-f cfgfile] [-b btnfile] "\
	  "[-F <font>] [-v] [-s] [-n]\n",app_name);
  fprintf(stderr,"\n");
  fprintf(stderr," wmbutton Ver %d Rel %d\n",VER,REL);
  fprintf(stderr,"\n");
  fprintf(stderr,"-g <geometry>  Window Geometry - ie: 64x64+10+10\n");
  fprintf(stderr,"-d <display>   Display -  ie: 127.0.0.1:0.0\n"); 
  fprintf(stderr,"-f <filename>  Full path to configuration file.\n");
  fprintf(stderr,"-b <filename>  Full path to button xpm.\n");
  fprintf(stderr,"-F <font>      Custom tooltip font (e.g. -b\\&h-lucidatypewriter-medium-*-*-*-12-*)\n");
  fprintf(stderr,"-v             Verbose Mode.\n");
  fprintf(stderr,"-h             Help. This message.\n");
#if MIDMOUSE
  fprintf(stderr,"-m             Disable Middle Mouse functionality.\n");
#else
  fprintf(stderr,"-m             Enable Middle Mouse functionality.\n");
#endif
  fprintf(stderr,"-s             Swap tooltip colors.\n");
  fprintf(stderr,"-n             Turn off tooltips.\n");
  fprintf(stderr,"\n");
  exit(0);
}/***********************************************************************/


/******  Error Handler Routine  *****************************************/
void err_mess(int err, char *str) {

  switch (err) {
  case FAILDISP:
    fprintf(stderr,"Fail: XOpenDisplay for %s\n", str);	
    exit(err);
  case FAILSWIN:
    fprintf(stderr,"Fail: XCreateSimpleWindow\n");	
    exit(err);
  case FAILICON: 
    fprintf(stderr,"Fail: XCreateSimpleWindow\n");	
    exit(err);
  case FAILXPM:
    fprintf(stderr,"Fail: XCreateBitmapFromData\n");
    break;
  case FAILWNAM:
    fprintf(stderr,"%s: Can't set up window name\n", str);
    exit(err);
  case FAILGC:
    fprintf(stderr,"Fail: XCreateGC\n");	
    exit(err);
  case FAILCONF:
    fprintf(stderr, "Fail: Can't Find configuration file %s\n",str);
    exit(err);
  case FAILTMPL:
    fprintf(stderr, "Fail: Can't Create 'template' Pixmap\n");
    exit(err);
  case FAILVIS:
    fprintf(stderr, "Fail: Can't Create 'visible' Pixmap\n");
    exit(err);
  case FAILBUT:
    fprintf(stderr, "Fail: Can't Create 'buttons' Pixmap\n");
    exit(err);
  default:
    fprintf(stderr, "Fail:  UnSpecified Error: %d\n",err);
    fprintf(stderr, "Fail:  %s\n",str);
    exit(err);
  }

}/***********************************************************************/

/***********************************************************************
 * RunAppN(int app)
 *
 * Run the command given in the configuration file 'configfile'
 ***********************************************************************/
void RunAppN( int app ) {
  char *cmndstr;
  extern int Verbose;

  cmndstr = Parse(app);                  // Get command to pass to system

  if (Verbose) fprintf(stderr, "Command String: %s", cmndstr);

  if (cmndstr != NULL) {
    system(cmndstr);  // if there's a command, run it
    free(cmndstr);
  }
}/***********************************************************************/

/***********************************************************************
 * Parse(int app)
 *
 * Parses the file 'configfile' for command to execute.
 ***********************************************************************/
char *Parse(int app) {
  FILE *fp;
  extern char configfile[FILEMAXLEN];
  char *ptr  = NULL;
  char *cmnd = NULL;
  char *line = NULL;

  if ((fp = fopen(configfile, "r")) == NULL)
    err_mess(FAILCONF,configfile);

  do {                               // Read Lines in config file
    line = readln(fp);
    if (line == NULL) break;         // if end of file, quit
    if ( (line[0] != '#') && (line[0] != '\n'))  // Ignore comments and blanks 
      if ( atoi(line) == app) break; // if we've found our app, quit looking
  } while (1);

  if (line == NULL) return(NULL);    // if we didn't find it, then just ignore

  ptr = strchr(line,'\t');                      // find first tab
  if (ptr == NULL)   ptr = strchr(line,' ');    // or space charater
  if (ptr == NULL)   return(NULL);              // if neither are there, its a badly
                                                // formatted line.  Ignore
  ptr++;
  cmnd = malloc((strlen(ptr)+1)*sizeof(char));  

  strcpy(cmnd, ptr);                 // Everything after number is passed to system

  free(line);
  fclose(fp);
  return(cmnd);
}/**********************************************************************/

/******  Function to Read a Line from fp  *******************************/
char *readln(FILE *fp){
  char *tmp, *t1, *t2;
  int count = 0, s1 = SIZE1;
  
  if((tmp = malloc(SIZE1*sizeof(char))) == NULL)return NULL;
  /* If cannot allocate memory */
  if((t1 = fgets(tmp, SIZE1, fp)) == NULL){
    free(tmp);
    return NULL;
  }				/* If there is no string to be read */
  while(strchr(t1,EOLN) == NULL){ /* Loop while EOLN is not in the string */
    count += strlen(t1);
    if((s1 - count) > 1)break;
    s1 += SIZE1;
    if((t2 = realloc(tmp, s1*sizeof(char))) == NULL){
      free(tmp);
      return NULL;
    }				/* If cannot allocate more memory */
    else tmp = t2;
    if((t1 = fgets((tmp+count),((s1 - count)),fp)) == NULL)break;
  }				/* End of While Loop */
  if((t2 = realloc(tmp,strlen(tmp)+1)) == NULL)free(tmp);
  return (t2);
}/**********************************************************************/


/***********************************************************************
 *   initTime
 *
 *     Copyright (c) 2001 Bruno Essmann <essmann@users.sourceforge.net>
 ***********************************************************************/

void initTime () {
  extern int Verbose;

  if (Verbose) {
    fprintf(stdout, "[        ] initializing time\n");
  }

  gettimeofday(&_tStart, NULL);
}/**********************************************************************/

long currentTimeMillis () {
  struct timeval tNow;
  struct timeval tElapsed;
  
  gettimeofday(&tNow, NULL);
  
  if (_tStart.tv_usec > tNow.tv_usec) {
    tNow.tv_usec+= 1000000;
    tNow.tv_sec--;
  }
  tElapsed.tv_sec= tNow.tv_sec - _tStart.tv_sec;
  tElapsed.tv_usec= tNow.tv_usec - _tStart.tv_usec;
  return (tElapsed.tv_sec * 1000) + (tElapsed.tv_usec / 1000);
}/**********************************************************************/


void getWindowOrigin (Window w, int* nX, int* nY) {
  extern Display *display;
  Window wWindow, wParent, wRoot;
  Window* wChildren;
  unsigned int nChildren;
  unsigned int ww, wh, wb, wd;
  int wx, wy;

  wParent= w;
  do {
    wWindow= wParent;
    if (!XQueryTree(display, wParent, &wRoot, &wParent, &wChildren, &nChildren))
      return;

    if (wChildren)
      XFree(wChildren);

  } while (wParent != wRoot);

  if (XGetGeometry(display, wWindow, &wRoot, &wx, &wy, &ww, &wh, &wb, &wd)) {
    if (nX) {
      *nX= wx;
    }
    if (nY) {
      *nY= wy;
    }
  }
}/**********************************************************************/

/***********************************************************************
 *  getButtonLocation
 *
 *  compute location for each button's tooltip (not perfect)
 ***********************************************************************/
void getButtonLocation (int nButton, int* nLocationX, int* nLocationY) {
  *nLocationX = 0;
  *nLocationY = 8;
  while (nButton > BUTTON_COLS) {
    *nLocationY += BUTTON_SIZE;
    nButton -= BUTTON_COLS;
  }
  while (nButton > 0) {
    *nLocationX += BUTTON_SIZE - 1;
    nButton--;
  }
}/**********************************************************************/

/***********************************************************************
 *  getNicenedString
 *
 *  nicen the parsed command from the .wmbutton config file
 *    - cut if too long
 *    - remove parameters, whitespace and the '&'...
 ***********************************************************************/
char* getNicenedString (char *old, int andAddSeparator) {
  int i = 0;
  char *new;
  char *tmp = NULL;

  new = (char*) malloc (33 * sizeof(char) + 8);
    
  if( old == NULL ) {
    strcpy( new, "--" );
    if( andAddSeparator )
      strcat( new, "  |  " );
    return new;
  }
  else if( strlen(old) > 33 ) {
    strncpy( new, old, 29 );
    new[30] = '\0';
    strcat( new, "..." );
  }
  else {
    strcpy( new, old );
  }

  if( (tmp = strstr( new, " -" )) != NULL ) {
    // cut off parameters...
    tmp[0] = '\0';
  }

  if( (i = strlen(new)) > 1 ) {
    // cut off whitespace and '&'
    while( isspace(new[i]) || iscntrl(new[i]) || new[i] == '&' )
      i--;
    
    if( i < strlen(new) )
      new[i+1] = '\0';
  }
  else {
    strcpy( new, "--" );
  }

  if( andAddSeparator )
    strcat( new, "  |  " );

  return(new);
}/**********************************************************************/

/***********************************************************************
 *  getButtonAppNames
 *
 *  returns the 1..3 application names / commands to be shown in tooltip
 ***********************************************************************/
char* getButtonAppNames (int nButton) {
  char *str = NULL;
  char *tmp;
    
  if (!( nButton < 0 || nButton > 9 )) {

    str = (char*) calloc (sizeof(char), 1024);

    tmp = getNicenedString( Parse( nButton + LMASK ),  1 );
    strcat( str, tmp );
    free(tmp);
    
    tmp = getNicenedString( Parse( nButton + MMASK ),  1 );
    strcat( str, tmp );
    free(tmp);
    
    tmp = getNicenedString( Parse( nButton + RMASK ),  0 );
    strcat( str, tmp );
    free(tmp);
  }

  return(str);
}/**********************************************************************/


int hasTooltipSupport () {
  return _bTooltipSupport;
}/**********************************************************************/

void showTooltip (int nButton, int nMouseX, int nMouseY) {
  Pixmap pixmap, mask;
  int nMainWinX, nMainWinY;
  int nButtonX = 0, nButtonY = 0, nButtonWidth = 0, nButtonHeight = 0;
  int nTextY, nX, nY, nWidth, nHeight, nSide;
  char* szText;
  extern int Verbose;
  extern Window iconwin;
  extern Pixel bg_pixel, fg_pixel;
  extern Display *display;
  extern GC gc;

  if (!_bTooltipSupport  ||  nButton == -1) {
    return;
  }
  if (_bTooltip) {
    hideTooltip();
  }


  if (Verbose) { 
    fprintf(stdout, "[%8ld] showing tooltip for button %d at %d, %d\n", 
	    currentTimeMillis(),
	    nButton, nMouseX, nMouseY);
  } 


  if( (szText = getButtonAppNames(nButton)) == NULL )
    return;

  szText = strdup(szText);

  _bTooltip= 1;

  nWidth= XTextWidth(_fTooltip, szText, strlen(szText)) + 16;
  nHeight= _nFontHeight + 4;
  if (nHeight < 16) {
    nHeight= 16;
  }
  if (nWidth < nHeight) {
    nWidth= nHeight;
  }
  if (Verbose) { 
    fprintf(stdout, "[%8ld] tooltip size: %d, %d\n", 
	    currentTimeMillis(), nWidth, nHeight);
  }
	
  getWindowOrigin(iconwin, &nMainWinX, &nMainWinY);
  getButtonLocation(nButton, &nButtonX, &nButtonY);
  nButtonX+= nMainWinX;
  nButtonY+= nMainWinY;
  nButtonWidth  = BUTTON_SIZE;
  nButtonHeight = BUTTON_SIZE;
	
  if (nButtonX + nWidth > _nScreenWidth) {
    nSide= TOOLTIP_RIGHT;
    nX= nButtonX - nWidth + nButtonWidth / 2;
    if (nX < 0) {
      nX= 0;
    }
  } else {
    nSide= TOOLTIP_LEFT;
    nX= nButtonX + nButtonWidth / 2;
  }
  if (nX + nWidth > _nScreenWidth) {
    nX= _nScreenWidth - nWidth;
  }

  if (nButtonY - (nHeight + TOOLTIP_SPACE) < 0) {
    nSide|= TOOLTIP_TOP;
    nY= nButtonY + nButtonHeight - 1;
    nTextY= TOOLTIP_SPACE;
  } else {
    nSide|= TOOLTIP_BOTTOM;
    nY= nButtonY - (nHeight + TOOLTIP_SPACE);
    nTextY= 0;
  }
	
  pixmap= createTooltipPixmap(nWidth, nHeight, nSide, &mask);

  XSetForeground(display, gc, _bTooltipSwapColors ? fg_pixel : bg_pixel);
  XSetFont(display, gc, _fTooltip->fid);
  XDrawString(display, pixmap, gc, 
	      8, nTextY + (nHeight - _nFontHeight) / 2 + _nFontY, 
	      szText, strlen(szText));
		
  XSetWindowBackgroundPixmap(display, _wTooltip, pixmap);

  XResizeWindow(display, _wTooltip, nWidth, nHeight + TOOLTIP_SPACE);
  XShapeCombineMask(display, _wTooltip, ShapeBounding, 0, 0, mask, ShapeSet);
  XFreePixmap(display, mask);
  XMoveWindow(display, _wTooltip, nX, nY);
  XMapRaised(display, _wTooltip);
  XFreePixmap(display, pixmap);

  free(szText);
}/**********************************************************************/

void hideTooltip () {
  extern int Verbose;
  extern Display *display;

  if (!_bTooltipSupport) {
    return;
  }
  if (_bTooltip) {
    if (Verbose) {
      fprintf(stdout, "[%8ld] hiding tooltip\n", currentTimeMillis());
    }
    XUnmapWindow(display, _wTooltip);
    _bTooltip= 0;
  }
}/**********************************************************************/

int hasTooltip () {
  if (!_bTooltipSupport) {
    return 0;
  }
  return _bTooltip;
}/**********************************************************************/

void initTooltip (int bTooltipSupport, char* szFontName, int bTooltipSwapColors) {
  XSetWindowAttributes attribs;
  unsigned long vmask;
  extern int Verbose;
  extern Display *display;
  extern char *app_name;
  extern int screen;
  extern Window rootwin, win;

  _bTooltipSwapColors = bTooltipSwapColors;
  _bTooltipSupport = bTooltipSupport;

  if (!_bTooltipSupport) {
    if (Verbose) {
      fprintf(stdout, "[%8ld] initializing tooltips (disabled)\n", 
	      currentTimeMillis());
    }

    return;
  }
  if (Verbose) {
    fprintf(stdout, "[%8ld] initializing tooltips\n", currentTimeMillis());
  }
  _fTooltip= XLoadQueryFont(display, szFontName);
  if (!_fTooltip) {
    fprintf(stderr, "%s: couldn't allocate font '%s'.\n", app_name, szFontName);
    if (!strcmp(szFontName, TOOLTIP_FONT))
      fprintf(stderr, "%s: Use option -F <font>\n", app_name);
    exit(-1);
  }
  _nFontHeight= _fTooltip->ascent + _fTooltip->descent;
  _nFontY= _fTooltip->ascent;
  _nScreenWidth= WidthOfScreen(ScreenOfDisplay(display, screen));
  _nScreenHeight= HeightOfScreen(ScreenOfDisplay(display, screen));
  if (Verbose) {
    fprintf(stdout, "[%8ld] configuring tooltip font:\n" \
	    "[%8ld] - '%s'\n" \
	    "[%8ld] - font-height= %d, font-ascent= %d\n" \
	    "[%8ld] configuring screen size: %dx%d\n",
	    currentTimeMillis(), 
	    currentTimeMillis(), szFontName, 
	    currentTimeMillis(), _nFontHeight, _nFontY,
	    currentTimeMillis(), _nScreenWidth, _nScreenHeight  );
  }

  vmask= CWSaveUnder | CWOverrideRedirect | CWBorderPixel;
  attribs.save_under= True;
  attribs.override_redirect= True;
  attribs.border_pixel= 0;
  _wTooltip= XCreateWindow(display, rootwin, 1, 1, 10, 10, 1,
			   CopyFromParent, CopyFromParent, 
			   CopyFromParent, vmask, &attribs);
  if (win == 0) {
    fprintf(stderr, "Cannot create tooltip window.\n");
    exit(-1);
  }
}/**********************************************************************/

void destroyTooltip () {
  extern Display *display;

  if (!_bTooltipSupport) {
    return;
  }
  if (_gcMono) {
    XFreeGC(display, _gcMono);
    _gcMono= 0;
  }
  XDestroyWindow(display, _wTooltip);
}/**********************************************************************/

void drawTooltipBalloon (Pixmap pix, GC gc, int x, int y, int w, int h, int side) {
  extern Display *display;
  int rad = h*3/10;
  XPoint pt[3];

  XFillArc(display, pix, gc, x, y, rad, rad, 90*64, 90*64);
  XFillArc(display, pix, gc, x, y+h-1-rad, rad, rad, 180*64, 90*64);

  XFillArc(display, pix, gc, x+w-1-rad, y, rad, rad, 0*64, 90*64);
  XFillArc(display, pix, gc, x+w-1-rad, y+h-1-rad, rad, rad, 270*64, 90*64);

  XFillRectangle(display, pix, gc, x, y+rad/2, w, h-rad);
  XFillRectangle(display, pix, gc, x+rad/2, y, w-rad, h);

  if (side & TOOLTIP_BOTTOM) {
    pt[0].y = y+h-1;
    pt[1].y = y+h-1+TOOLTIP_SPACE;
    pt[2].y = y+h-1;
  } else {
    pt[0].y = y;
    pt[1].y = y-TOOLTIP_SPACE;
    pt[2].y = y;
  }
  if (side & TOOLTIP_RIGHT) {
    pt[0].x = x+w-h+2*h/16;
    pt[1].x = x+w-h+11*h/16;
    pt[2].x = x+w-h+7*h/16;
  } else {
    pt[0].x = x+h-2*h/16;
    pt[1].x = x+h-11*h/16;
    pt[2].x = x+h-7*h/16;
  }
  XFillPolygon(display, pix, gc, pt, 3, Convex, CoordModeOrigin);
}/**********************************************************************/

Pixmap createTooltipPixmap (int width, int height, int side, Pixmap *mask) {
  extern Display *display;
  extern GC gc;
  extern Pixel bg_pixel, fg_pixel;
  extern int depth;
  extern Window rootwin;
  Pixmap bitmap;
  Pixmap pixmap;
  int x, y;

  bitmap = XCreatePixmap(display, rootwin, 
			 width+TOOLTIP_SPACE, height+TOOLTIP_SPACE, 1);

  if (!_gcMono) {
    _gcMono= XCreateGC(display, bitmap, 0, NULL);
  }
  XSetForeground(display, _gcMono, 0); 
  XFillRectangle(display, bitmap, _gcMono, 0, 0, 
		 width+TOOLTIP_SPACE, height+TOOLTIP_SPACE);

  pixmap = XCreatePixmap(display, rootwin, 
			 width+TOOLTIP_SPACE, height+TOOLTIP_SPACE, depth);
  XSetForeground(display, gc, _bTooltipSwapColors ? fg_pixel : bg_pixel);
  XFillRectangle(display, pixmap, gc, 0, 0, 
		 width+TOOLTIP_SPACE, height+TOOLTIP_SPACE);

  if (side & TOOLTIP_BOTTOM) {
    y = 0;
  } else {
    y = TOOLTIP_SPACE;
  }
  x = 0;

  XSetForeground(display, _gcMono, 1);
  drawTooltipBalloon(bitmap, _gcMono, x, y, width, height, side);
  XSetForeground(display, gc, _bTooltipSwapColors ? bg_pixel : fg_pixel);
  drawTooltipBalloon(pixmap, gc, x+1, y+1, width-2, height-2, side);

  *mask = bitmap;

  return pixmap;
}/***********************************************************************/


/***********************************************************************
 *   flush_expose
 *
 *   Everyone else has one of these... Can't hurt to throw it in.
 ***********************************************************************/
int flush_expose(Window w) {
  extern Display *display;
  XEvent         dummy;
  int            i=0;

  while (XCheckTypedWindowEvent(display, w, Expose, &dummy)) i++;
  return(i);
}/***********************************************************************/
